using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace TaskScheduling
{
    public partial class frmTaskScheduling : Form
    {
        #region Members
        Int32 iTaskCount;
        Int32 iProcessorCount;
        enum SortType { Ascending = 0, Descending = 1 };
        private int[,] iMatrix;
        private int[] Processor;
        #endregion

        #region Initialize
        public frmTaskScheduling()
        {
            InitializeComponent();
        }
        #endregion

        #region Methods
        //set Grid view be andaze tasks
        private void SetGrid()
        {
            try
            {
                int i = 0;
                grdTaskTime.ClearItems();
                if (Convert.ToInt32(txtTaskCount.Text) / 10 > 0)
                    for (i = 0; i < Convert.ToInt32(txtTaskCount.Text) / 10; i++)
                        grdTaskTime.AddItem(i);
                if (Convert.ToInt32(txtTaskCount.Text) % 10 > 0)
                    grdTaskTime.AddItem(i);
            }
            catch
            {
            }
        }

        private void Swap(int i, int j)
        {
            int temp;

            temp = iMatrix[0, i];
            iMatrix[0, i] = iMatrix[0, j];
            iMatrix[0, j] = temp;

            temp = iMatrix[1, i];
            iMatrix[1, i] = iMatrix[1, j];
            iMatrix[1, j] = temp;
        }

        //sort tasks
        private void SortArray(int SortTypeFlag,int StartIndex, int EndIndex)
        {
            for (int i = StartIndex; i < EndIndex; i++)
                for (int j = i + 1; j < EndIndex; j++)
                    if (SortTypeFlag == (int)SortType.Ascending)
                    {
                        if (iMatrix[1, i] > iMatrix[1, j])
                            Swap(i,j);
                    }
                    else
                    {
                        if (iMatrix[1, i] < iMatrix[1, j])
                            Swap(i,j);
                    }
        }

        private bool ReadData()
        {
            try
            {                
                int i = 0, j = 0, c = 0;
                //Check Input Data
                if (txtTaskCount.Text == null || txtTaskCount.Text == "")
                {
                    MessageBox.Show("Enter Task Count.","TaskScheduling",MessageBoxButtons.OK,MessageBoxIcon.Error);
                    txtTaskCount.Focus();
                    return false;
                }
                if (txtProcessCount.Text == null || txtProcessCount.Text == "")
                {
                    MessageBox.Show("Enter Processor Count.", "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    txtProcessCount.Focus();
                    return false;
                }
                //Set Task Count & Processor Count
                iTaskCount = Convert.ToInt32(txtTaskCount.Text);
                iProcessorCount = Convert.ToInt32(txtProcessCount.Text);
                //Set Processor Array
                Processor = new int[iProcessorCount];
                //Set Task ExecutionTime Array
                iMatrix = new int[4, iTaskCount];

                //?
                while (c < iTaskCount)
                    {
                        for (i = 0; i < iTaskCount / 10; i++)
                        {
                            grdTaskTime.Row = i;
                            for (j = 0; j < 10; j++)
                            {
                                iMatrix[0, c] = c;
                                iMatrix[1, c++] = Convert.ToInt32(grdTaskTime.GetRow().Cells[j].Value);
                            }
                        }
                        if (iTaskCount % 10 > 0)
                        {
                            grdTaskTime.Row = i;
                            for (i = 0; i < iTaskCount % 10; i++)
                            {
                                iMatrix[0, c] = c;
                                iMatrix[1, c++] = Convert.ToInt32(grdTaskTime.GetRow().Cells[i].Value);
                            }
                        }
                    }
                    return true;
            }
            catch (Exception exp)
            {
                MessageBox.Show(exp.Message, "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return false;
            }
        }
        //?
        private int FindFreeProcessor()
        {
            int min = 0;
            int minValue = Processor[0];
            for (int i = 1; i < iProcessorCount; i++)
                if (Processor[i] < minValue)
                {
                    minValue = Processor[i];
                    min = i;
                }
            return min;
        }

        // Calculate Finishing Time
        private int CalculateFinishingTime()
        {
            try
            {
                int maxValue = Processor[0];
                for (int i = 1; i < iProcessorCount; i++)
                    if (Processor[i] > maxValue)
                        maxValue = Processor[i];
                return maxValue;
            }
            catch(Exception exp)
            {
                MessageBox.Show(exp.Message, "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return 0;
            }
        }

        //Calculate Waiting Time
        private decimal CalculateWaitingTime()
        {
            try
            {
                int i, j, w;
                for (i = 0; i < iProcessorCount; i++)
                {
                    w = 0;
                    for (j = 0; j < iTaskCount; j++)
                        if (iMatrix[2, j] == i)
                        {
                            iMatrix[3, j] = w;
                            w += iMatrix[1, j];
                        }
                }

                w = 0;
                for (i = 0; i < iTaskCount; i++)
                    w += iMatrix[3, i];
                return Math.Round((decimal)w / iTaskCount, 2);
            }
            catch (Exception exp)
            {
                MessageBox.Show(exp.Message,"TaskScheduling", MessageBoxButtons.OK,MessageBoxIcon.Error);
                return 0;
            }
        }

        //Print Output
        private void PrintOutput()
        {
            int i, j, k;
            grdOutput.RootTable.Columns.Clear();
           
            for (i = 0; i < iProcessorCount; i++)
            {
                grdOutput.RootTable.Columns.Add("P" + i.ToString());
                grdOutput.RootTable.Columns[i].Width = 70;
                grdOutput.RootTable.Columns[i].AllowSort = false;
                grdOutput.RootTable.Columns[i].TextAlignment = Janus.Windows.GridEX.TextAlignment.Center;
                grdOutput.RootTable.Columns[i].BoundMode = Janus.Windows.GridEX.ColumnBoundMode.Unbound;
                grdOutput.RootTable.Columns[i].SelectableCells = Janus.Windows.GridEX.SelectableCells.None;
            }
            grdOutput.HeaderFormatStyle.TextAlignment = Janus.Windows.GridEX.TextAlignment.Center;
            //Print Output  
            i = grdOutput.RowCount -1;
            while (i>-1)
            {
                grdOutput.Row = 0;
                grdOutput.CurrentRow.Delete();
                i--;
            }
            for (i = 0; i < iTaskCount; i++)
                grdOutput.AddItem(i);

            for (i = 0; i < iProcessorCount; i++)
            {
                k = -1;
                for (j = 0; j < iTaskCount; j++)
                {
                    if (iMatrix[2, j] == i)
                    {                        
                        grdOutput.Row = ++k;
                        grdOutput.CurrentRow.BeginEdit();
                        grdOutput.CurrentRow.Cells["P" + i].Value = "T" + iMatrix[0, j] + " (" + iMatrix[1, j].ToString() + ")";
                        grdOutput.CurrentRow.Cells["P" + i].Text = "T" + iMatrix[0, j] + " (" + iMatrix[1, j].ToString() + ")";
                        grdOutput.CurrentRow.EndEdit();
                    }
                }
            }
            lblFinishingTime.Text = CalculateFinishingTime().ToString();
            lblWaitingTime.Text = CalculateWaitingTime().ToString();
        }
        #endregion    

        #region Events
        private void frmTaskScheduling_Load(object sender, EventArgs e)
        {
            grdTaskTime.RootTable.Columns.Clear();
            for (int i = 0; i < 10; i++)
            {
                grdTaskTime.RootTable.Columns.Add(i.ToString());
                grdTaskTime.RootTable.Columns[i].Width = 50;
                grdTaskTime.RootTable.Columns[i].InputMask = "General0";
                grdTaskTime.RootTable.Columns[i].TextAlignment = Janus.Windows.GridEX.TextAlignment.Center;
                grdTaskTime.RootTable.Columns[i].BoundMode = Janus.Windows.GridEX.ColumnBoundMode.Unbound;
            }
            grdTaskTime.RootTable.ColumnHeaders = Janus.Windows.GridEX.InheritableBoolean.False;

            txtStartRandomNumber.Text = "1";
            txtEndRandomNumber.Text = "100";
            txtAlpha.Text = "100";
        }
       
        //
        private void btnRandom_Click(object sender, EventArgs e)
        {
            try
            {
                //Check Input Data
                if (txtTaskCount.Text == null || txtTaskCount.Text == "")
                {
                    MessageBox.Show("Enter Task Count.", "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    txtTaskCount.Focus();
                    return;
                }
                if (txtProcessCount.Text == null || txtProcessCount.Text == "")
                {
                    MessageBox.Show("Enter Processor Count.", "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    txtProcessCount.Focus();
                    return;
                }
                //Set Task Count & Processor Count
                iTaskCount = Convert.ToInt32(txtTaskCount.Text);
                iProcessorCount = Convert.ToInt32(txtProcessCount.Text);
                //Set Random Number Limitation
                if (txtStartRandomNumber.Text == null || txtStartRandomNumber.Text == "")
                    txtStartRandomNumber.Text = "1";
                if (txtEndRandomNumber.Text == null || txtEndRandomNumber.Text == "")
                    txtEndRandomNumber.Text = "100";
                //Generate Random Numbers
                int i, j;
                Random rnd = new Random();

                grdTaskTime.ClearItems();

                for (i = 0; i < iTaskCount / 10; i++)
                {
                    grdTaskTime.AddItem(i);
                    grdTaskTime.Row = i;
                    grdTaskTime.CurrentRow.BeginEdit();
                    for (j = 0; j < 10; j++)
                    {
                        grdTaskTime.CurrentRow.Cells[j].Value = rnd.Next(Convert.ToInt32(txtStartRandomNumber.Text), Convert.ToInt32(txtEndRandomNumber.Text));
                        grdTaskTime.CurrentRow.Cells[j].Text = rnd.Next(Convert.ToInt32(txtStartRandomNumber.Text), Convert.ToInt32(txtEndRandomNumber.Text)).ToString();
                    }
                    grdTaskTime.CurrentRow.EndEdit();
                }
                if (iTaskCount % 10 > 0)
                {
                    grdTaskTime.AddItem(i);
                    grdTaskTime.Row = i;
                    grdTaskTime.CurrentRow.BeginEdit();
                    for (i = 0; i < iTaskCount % 10; i++)
                    {
                        grdTaskTime.CurrentRow.Cells[i].Value = rnd.Next(Convert.ToInt32(txtStartRandomNumber.Text), Convert.ToInt32(txtEndRandomNumber.Text));
                        grdTaskTime.CurrentRow.Cells[i].Text = rnd.Next(Convert.ToInt32(txtStartRandomNumber.Text), Convert.ToInt32(txtEndRandomNumber.Text)).ToString();
                    }
                    grdTaskTime.CurrentRow.EndEdit();
                }
            }
            catch (Exception exp)
            {
                MessageBox.Show(exp.Message, "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        // GeneticAlgorithm
        private void btnGeneticAlgorithm_Click(object sender, EventArgs e)
        {
           // this.Cursor = Cursors.WaitCursor;
            if (txtAlpha.Text == null || txtAlpha.Text == "") txtAlpha.Text = "100";
            if (txtGenerationCount.Text == null || txtGenerationCount.Text == "") txtGenerationCount.Text = "10";
            if (!GenerateInitialPopulation())
            {
                this.Cursor = Cursors.Default;
                return;
            }
            for (int i = 0; i < Convert.ToInt32(txtGenerationCount.Text); i++)
            {
                CrossOver();
                Mutation();
                CalculateProbability();
                CalculateAggregationProbability();
                Selection();
            }
            PrintBestSolution();
            this.Cursor = Cursors.Default;
        }        
        #endregion

        #region GeneticAlgorithmMethods
        //*****************************
        decimal Alpha;
        int iChromosomeCount;
        string[,] Chromosome;
        string[,] NextGeneration;
        int[] FinishingTime;
        decimal[] WaitingTime;
        decimal[] Fitness;
        decimal[] Probability;
        decimal[] AggregationProbability;
        double[] SelectionNumbers;
        int[] SelectedChromosome;
        int[] CountOfSelectedChromosomes;
        //*****************************
        private string ConvertToBinaryNumber(int iNumber, int iBitCounts)
        {
            string str = "", rstr = "";
            while (iNumber >= 2)
            {
                str += iNumber % 2;
                iNumber /= 2;
            }
            str += iNumber.ToString();
            for (int i = str.Length -1; i >=0; i--)
                rstr += str[i];
            str = rstr;
            if (iBitCounts != 0)
                str = str.PadLeft(iBitCounts, '0');
            return str;
        }

        private int ConvertToDecimalNumber(string str)
        {
            double iResult = 0, j = -1;
            for (int i = str.Length - 1; i>=0; i--)
                iResult += Math.Pow(2, ++j) * Convert.ToDouble(str[i].ToString());
            return (int)iResult;
        }

        private bool GenerateInitialPopulation()
        {
            if (!ReadData()) return false;
            if (txtChromosomeCount.Text == null || txtChromosomeCount.Text == "")
            {
                MessageBox.Show("Enter Chromosome Count.", "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
                txtChromosomeCount.Focus();
                return false;                
            }
            //***********************************            
            int iBitCounts = 0;
            Random rnd = new Random();            
            iBitCounts = ConvertToBinaryNumber(iProcessorCount - 1, 0).Length;

            iChromosomeCount = Convert.ToInt32(txtChromosomeCount.Text);
            Chromosome = new string[iChromosomeCount, iTaskCount];
            for (int i = 0; i < iChromosomeCount; i++)
            {
                for (int j = 0; j < iTaskCount; j++)
                {
                    int p = rnd.Next(0, iProcessorCount);
                    Chromosome[i, j] = ConvertToBinaryNumber(p, iBitCounts);
                }
            }
            return true;
        }

        private decimal CalculateWaitingTime_GA(int ChromosomeNumber)
        {
            try
            {
                int i, j, w;
                for (i = 0; i < iProcessorCount; i++)
                {
                    w = 0;
                    for (j = 0; j < iTaskCount; j++)
                        if (ConvertToDecimalNumber(Chromosome[ChromosomeNumber, j]) == i)
                        {
                            iMatrix[3, j] = w;
                            w += iMatrix[1, j];
                        }
                }

                w = 0;
                for (i = 0; i < iTaskCount; i++)
                    w += iMatrix[3, i];
                return Math.Round((decimal)w / iTaskCount, 2);
            }
            catch (Exception exp)
            {
                MessageBox.Show(exp.Message, "TaskScheduling", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return 0;
            }
        }

        private void CalculateFinishingAndWaitingTime()
        {
            FinishingTime = new int[iChromosomeCount];
            WaitingTime = new decimal[iChromosomeCount];        

            for (int i = 0; i < iChromosomeCount; i++)
            {
                Processor = new int[iProcessorCount];
                for (int j = 0; j < iTaskCount; j++)
                {
                    Processor[ConvertToDecimalNumber(Chromosome[i, j])] += iMatrix[1,j];                    
                }
                FinishingTime[i] = CalculateFinishingTime();
                WaitingTime[i] = CalculateWaitingTime_GA(i);
            }
        }        

        private void CalculateFitness()
        {
            Alpha = Convert.ToByte(txtAlpha.Text);
            Fitness = new decimal[iChromosomeCount];
            for (int i = 0; i < iChromosomeCount; i++)
                Fitness[i] = 1 / ((decimal)(Alpha / 100) * FinishingTime[i] + (1 - Alpha / 100) * WaitingTime[i]);
        }

        private bool Exists(int[] random, int r)
        {
            for (int i = 0; i < iChromosomeCount; i++)
                if (random[i] == r)
                    return true;
            return false;
        }

        private void CrossOver()
        {
            int i, j, r;
            Random rnd = new Random();
            int[] random = new int[iChromosomeCount];
            for (i = 0; i < iChromosomeCount; i++) random[i] = -1;
            //Generate random numbers
            for (i = 0; i < iChromosomeCount; i++)
            {
                r = rnd.Next(0, iChromosomeCount);
                while (Exists(random, r))
                {
                    r = rnd.Next(0, iChromosomeCount);
                }
                random[i] = r;
            }
            //*************************
            int MinPoint = rnd.Next(0,iTaskCount);
            int MaxPoint = rnd.Next(0, iTaskCount);
            if (MinPoint > MaxPoint)
            {
                MinPoint += MaxPoint;
                MaxPoint = MinPoint - MaxPoint;
                MinPoint = MinPoint - MaxPoint;
            }

            int MoveCount = iChromosomeCount % 2 != 0 ? iChromosomeCount - 1 : iChromosomeCount;
            for (i = 0; i < MoveCount; i+=2)
            {
                for (j = MinPoint; j <= MaxPoint; j++)
                {
                    string temp = Chromosome[random[i], j];
                    Chromosome[random[i], j] = Chromosome[random[i + 1 ], j];
                    Chromosome[random[i + 1], j] = temp;                    
                }
            }
        }

        private void Mutation()
        {
            Random rnd = new Random();
            double MutationRate = rnd.NextDouble();

            for (int i = 0; i < iChromosomeCount; i++)
            {
                for (int j = 0; j < iTaskCount; j++)
                {
                    do
                    {
                        string str = Chromosome[i, j];
                        for (int k = 0; k < str.Length; k++)
                        {
                            if (rnd.NextDouble() > MutationRate)
                            {
                                if (str[k] == '0')
                                {
                                    if (k == 0)
                                        str = "1" + str.Substring(k + 1, str.Length - k - 1);
                                    else
                                        str = str.Substring(0, k) + "1" + str.Substring(k + 1, str.Length - k -1);
                                }
                                else
                                {
                                    if (k == 0)
                                    {
                                        if (k + 1 <= str.Length - k - 1)
                                            str = "0" + str.Substring(k + 1, str.Length - k - 1);
                                        else
                                            str = "0";
                                    }
                                    else
                                        str = str.Substring(0, k) + "0" + str.Substring(k + 1, str.Length - k -1);
                                }
                            }
                        }
                        Chromosome[i, j] = str;
                    } while (ConvertToDecimalNumber(Chromosome[i, j]) > iProcessorCount - 1);
                }
            }
        }

        private decimal CalculateSumOfFitness()
        {
            decimal  SumOfFitness =0;
            for (int i = 0; i < iChromosomeCount; i++)
                SumOfFitness += Fitness[i];
            return SumOfFitness;
        }

        private void CalculateProbability()
        {
            Probability = new decimal[iChromosomeCount];
            CalculateFinishingAndWaitingTime();
            CalculateFitness();
            decimal SumOfFitness = CalculateSumOfFitness();
            for (int i = 0; i < iChromosomeCount; i++)
                Probability[i] = Fitness[i] / SumOfFitness;            
        }

        private void CalculateAggregationProbability()
        {
            AggregationProbability = new decimal[iChromosomeCount];
            AggregationProbability[0] = Probability[0];
            for (int i = 1; i < iChromosomeCount; i++)
                AggregationProbability[i] += AggregationProbability[i - 1] + Probability[i];
        }

        private void Selection()
        {
            //Generate random nummbers between 0, 1
            double r;
            Random rnd = new Random();
            SelectionNumbers = new double[iChromosomeCount];
            for (int i = 0; i < iChromosomeCount; i++)
            {
                r = rnd.NextDouble();
                while (r == 0)
                    r = rnd.NextDouble();
                SelectionNumbers[i] = r;
            }

            //Select Chromosomes
            SelectedChromosome = new int [iChromosomeCount];
            for (int i = 0; i < iChromosomeCount; i++)
            {
                int j=1;
                while ((double)AggregationProbability[j] < SelectionNumbers[i])
                    j++;
                SelectedChromosome[i] = j;
            }

            //CountOfSelectedChromosomes
            CountOfSelectedChromosomes = new int[iChromosomeCount];
            for (int i = 0; i < iChromosomeCount; i++)
                CountOfSelectedChromosomes[SelectedChromosome[i]]++;
            //Generate Next Generation
            NextGeneration = new string[iChromosomeCount, iTaskCount];
            int row = -1;
            for (int i = 0; i < iChromosomeCount; i++)
            {
                for (int j = 0; j < CountOfSelectedChromosomes[i]; j++)
                {
                    row++;
                    for (int k = 0; k < iTaskCount; k++)
                        NextGeneration[row, k] = Chromosome[i, k];
                }                    
            }
            Chromosome = NextGeneration;
        }

        private void PrintBestSolution()
        {
            int i, j, k;
            decimal BestFitness = Fitness[0];
            int BestChromosome = 0;

            for (i = 1; i < iChromosomeCount; i++)
            {
                if (Fitness[i] > BestFitness)
                {
                    BestFitness = Fitness[i];
                    BestChromosome = i;
                }
            }

            grdOutput.RootTable.Columns.Clear();
            for (i = 0; i < iProcessorCount; i++)
            {
                grdOutput.RootTable.Columns.Add("P" + i.ToString());
                grdOutput.RootTable.Columns[i].Width = 70;
                grdOutput.RootTable.Columns[i].AllowSort = false;
                grdOutput.RootTable.Columns[i].TextAlignment = Janus.Windows.GridEX.TextAlignment.Center;
                grdOutput.RootTable.Columns[i].BoundMode = Janus.Windows.GridEX.ColumnBoundMode.Unbound;
                grdOutput.RootTable.Columns[i].SelectableCells = Janus.Windows.GridEX.SelectableCells.None;
            }
            grdOutput.HeaderFormatStyle.TextAlignment = Janus.Windows.GridEX.TextAlignment.Center;
            //Print Output  
            i = grdOutput.RowCount - 1;
            while (i > -1)
            {
                grdOutput.Row = 0;
                grdOutput.CurrentRow.Delete();
                i--;
            }
            for (i = 0; i < iTaskCount; i++)
                grdOutput.AddItem(i);

            int FT = 0;
            for (i = 0; i < iProcessorCount; i++)
            {
                k = -1;
                int sum = 0;
                for (j = 0; j < iTaskCount; j++)
                {
                    if (ConvertToDecimalNumber(Chromosome[BestChromosome,j]) == i)
                    {
                        sum += iMatrix[1, j];
                        grdOutput.Row = ++k;
                        grdOutput.CurrentRow.BeginEdit();
                        grdOutput.CurrentRow.Cells["P" + i].Value = "T" + j.ToString() + " (" + iMatrix[1, j].ToString() + ")";
                        grdOutput.CurrentRow.Cells["P" + i].Text = "T" + j.ToString() +" (" + iMatrix[1, j].ToString() + ")";                        
                        grdOutput.CurrentRow.EndEdit();
                    }
                }
                if (FT < sum) FT = sum;
            }
            lblFinishingTime.Text = FT.ToString();
            lblWaitingTime.Text =Math.Round(CalculateWaitingTime_GA(BestChromosome),2).ToString();
            grdOutput.Row = 0;
        }
        #endregion      

        private void groupBox2_Enter(object sender, EventArgs e)
        {

        }
    }
}